/*
 *  ulocks.h   :  Fast User Level Synchronization 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Copyright (c) International Business Machines Corp., 2001
 *
 * Author: Hubertus Franke <frankeh@us.ibm.com, frankeh@watson.ibm.com>
 *
 */

#ifndef __ULOCKS_H__
#define __ULOCKS_H__

#include <linux/unistd.h>
#include <asm/atomic.h>
#include <linux/cache.h>

/* basic types supported */
#define ULOCK_SEMAPHORE  0
#define ULOCK_RWLOCK     1

/* basic operations supported */
#define ULOCK_OP_WAIT    0
#define ULOCK_OP_SIGNAL  1
#define ULOCK_OP_CREATE  2
#define ULOCK_OP_DESTROY 3

#define ULOCK_STAT_COUNTERS 5

typedef struct ulock_t {
	unsigned long  status;
	unsigned long  type;
	unsigned long  counters[ULOCK_STAT_COUNTERS];
	char           pad[SMP_CACHE_BYTES - 
			   (ULOCK_STAT_COUNTERS+2)*sizeof(unsigned long)];
} ulock_t ____cacheline_aligned;

/* UserLevel Semaphores == Fast Semaphores == Locks == Exclusive Locks */

#define INITIALIZED_USEMA(count) { count , ULOCK_SEMAPHORE }

/* UserLevel ReadWrite Locks == Shared Locks */

#define RW_WRITE_LOCKED          0x80000000
#define RW_WRITER_MASK           0x7FFF0000
#define RW_READER_MASK           0x0000FFFF
#define RW_WRITER_WAITING_INC    0x00010000


#define __NR_ulock_op  232

/**************************** kernel API *************************/

#ifndef __KERNEL__

#include <asm/ulocks.h>   // platform specific atomic ops

static inline _syscall3(int,ulock_op,int,op,void*,lword,long,arg);
#define ulock_wait(lword,arg)     ulock_op(ULOCK_OP_WAIT,lword,arg)
#define ulock_signal(lword,arg)   ulock_op(ULOCK_OP_SIGNAL,lword,arg)
#define ulock_create(lword)       ulock_op(ULOCK_OP_CREATE,lword,0)
#define ulock_destroy(lword)      ulock_op(ULOCK_OP_DESTROY,lword,0)

/************************* lock statistics counter  ****************/

#define ULOCK_STAT_WRITE_SUCCESS       0
#define ULOCK_STAT_WRITE_FAILED        1
#define ULOCK_STAT_READ_SUCCESS        2
#define ULOCK_STAT_READ_FAILED         3
#define ULOCK_STAT_WRITE_SPIN_SUCCESS  4

#ifdef LOCK_STATISTICS
#  define inc_stat_safe(ulock,cnt) u_atomic_inc(&(ulock)->counters[cnt])
#  define inc_stat(ulock,cnt)      ((ulock)->counters[cnt]++)
static inline 
void INIT_COUNTERS(ulock_t *ulock)
{ 
	int i;
	for (i=0;i<ULOCK_STAT_COUNTERS;i++) ulock->counters[i] = 0;
}
#else
#  define inc_stat_safe(ulock,cnt)
#  define inc_stat(ulock,cnt) 
   static inline void INIT_COUNTERS(ulock_t *ulock) { }
#endif

/*********************************************************************
 * UserLevel Semaphores == Locks == Exclusive Locks
 *********************************************************************/

static inline
void usema_init(ulock_t *ulock, int count)
{
	ulock->status = count;
	ulock->type   = ULOCK_SEMAPHORE;
	INIT_COUNTERS(ulock);
}

static inline 
int usema_down(ulock_t *ulock)
{
	if (!__ulock_down(ulock)) {
		inc_stat(ulock,ULOCK_STAT_WRITE_SUCCESS);
		return 0;
	}
	inc_stat_safe(ulock,ULOCK_STAT_WRITE_FAILED);
	return ulock_wait(ulock,0);
}

static inline 
int usema_up(ulock_t *ulock)
{
	if (!__ulock_up(ulock))
		return 0;
	return ulock_signal(ulock,0);
}

extern int __ulock_down_spin(ulock_t *, unsigned long);

static inline 
int usema_down_spin(ulock_t *ulock, long usecs)
{
	long status = (long)ulock->status;
	
	if ((status > 0) && 
	    (__compare_and_swap(&ulock->status,status,status-1) == status))
	{
		inc_stat(ulock,ULOCK_STAT_WRITE_SUCCESS);
		return 0;
	}
	inc_stat_safe(ulock,ULOCK_STAT_WRITE_FAILED);
	return __ulock_down_spin(ulock,usecs);
}

/*********************************************************************
 * UserLevel ReadWrite Locks aka Shared Locks
 * 
 *   ReadWrite Locks can take on various semantics on lock release
 *   (a)   reader preference (always start readers before writers)
 *   (b)   writer preference (always start writers before readers)
 *   (c)   fair (start in order of arrival)
 *********************************************************************/

#define CMP_SWAP(newval) 						        \
do {									        \
	if (__compare_and_swap(&ulock->status,old_status,newval)!=old_status)   \
		goto try_again;						        \
} while (0)


static inline 
int urwlock_rp_init(ulock_t *ulock)
{
	ulock->status = 0;
	ulock->type   = ULOCK_RWLOCK;
	INIT_COUNTERS(ulock);
	return 0;
}

static inline 
int urwlock_rp_readlock(ulock_t *ulock)
{
	if (!__urwlock_read_down(ulock)) {
		inc_stat_safe(ulock,ULOCK_STAT_READ_SUCCESS);
		return 0;
	}
	inc_stat_safe(ulock,ULOCK_STAT_READ_FAILED);
	return ulock_wait(ulock, 1);  /* read locked */
}

static inline 
int urwlock_rp_readunlock(ulock_t *ulock)
{
	unsigned long old_status;
	unsigned long status;

 try_again:
	old_status = ulock->status;
	status = old_status - 1;
	if ((status & RW_READER_MASK) ||         /* more readers */
	    ((status & RW_WRITER_MASK) == 0))    /* no pending writers */
	{
		CMP_SWAP (status);
		return 0;
	} else {
		CMP_SWAP ((status - RW_WRITER_WAITING_INC) | RW_WRITE_LOCKED);
		return ulock_signal(ulock,-1);
	}
}		


static inline 
int urwlock_rp_writelock(ulock_t *ulock)
{
	unsigned long old_status;

 try_again:
	old_status = __compare_and_swap(&ulock->status,0,RW_WRITE_LOCKED);
	if (old_status != 0) {
		CMP_SWAP( old_status + RW_WRITER_WAITING_INC );
		inc_stat_safe(ulock,ULOCK_STAT_WRITE_FAILED);
		return ulock_wait(ulock,0);
	}
	inc_stat(ulock,ULOCK_STAT_WRITE_SUCCESS);
	return 0;
}

static inline 
int urwlock_rp_writeunlock(ulock_t *ulock)
{
	unsigned long old_status;
	unsigned long rdcnt;

	/* fast uncontended path */
 try_again:
	old_status = ulock->status;
	rdcnt = old_status & RW_READER_MASK;
	if (rdcnt)  /* readers pending */
	{ 
		CMP_SWAP (old_status & ~RW_WRITE_LOCKED);
		ulock_signal(ulock,rdcnt); /* unlock readers */
		return 0;
	} 
	if (old_status & RW_WRITER_MASK)  /* pending writers */
	{
		CMP_SWAP(old_status-RW_WRITER_WAITING_INC);
		ulock_signal(ulock,-1); /* unlock writer */
		return 0;
	}

	CMP_SWAP ( old_status & ~RW_WRITE_LOCKED );
	return 0;
}

// default behavior is reader preference

#define urwlock_init        urwlock_rp_init
#define urwlock_readlock    urwlock_rp_readlock
#define urwlock_readunlock  urwlock_rp_readunlock
#define urwlock_writelock   urwlock_rp_writelock
#define urwlock_writeunlock urwlock_rp_writeunlock


// other behavior such as writer preference will be added later: HF

#endif
#endif
