/*
 * ulockflex.c - flexible benchmark  to test userlevel locks
 *
 * Copyright (C) 2001 IBM
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * 15 Oct 2001 - 1.0.0 Initial version
 * 12 Dec 2001 - 1.0.1 Clean up, process, thead introduction
 *                     statistics support
 *
 * Written by Hubertus Franke (frankeh@us.ibm.com) 15 Oct 2001
 *
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sched.h>
#include <errno.h>
#include <wait.h>
#include <math.h>
#include <getopt.h>
#include <stdarg.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/shm.h>
#include <asm/mman.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <errno.h>
#include <asm/atomic.h>

#include "setup.h"


/************************************************/

#define MAX_CHILDREN    (1024)
#define MAX_ULOCKS      (512)

#define STACK_SIZE	(8192)
#define CLONE_FLAGS_PROCESS	(CLONE_FS)
#define CLONE_FLAGS_THREAD      (CLONE_VM | CLONE_FS)
#define NUM_WARMUP	(num_warmup)
#define MIN_TRIALS	(NUM_WARMUP+5)
#define MAX_TRIALS	(25)

#include <linux/ulocks.h>

ulock_t *ulocks[MAX_ULOCKS];             // array to pointer of user locks

/***************************************************************************/


int num_warmup = 1;  /* default behavior */

#define HISTORY_COUNT (1<<15) 
#define HISTORY_IDX(x)  ((x) & ~(HISTORY_COUNT-1))

#define bprintf // printf   /* note this is not thread safe and hence deactivated */

struct child_data {
	unsigned long long count;
	int  pid;
	char pad[L1_CACHE_BYTES - sizeof(int) - sizeof(unsigned long long)];
};

struct lock_data {
	atomic_t verify_values;
	char pad[L1_CACHE_BYTES - sizeof(atomic_t) ];
};

struct shared_data {
	struct child_data per_child[MAX_CHILDREN];
	struct lock_data  per_lock[MAX_ULOCKS];

	int history[HISTORY_COUNT];  // wrap around buffer for recording
	atomic_t hcnt;               // index 

	int stop_test;		     // flag to end test when convergence met
};

struct shared_data  shared_global;
struct shared_data  *shared = &shared_global;

#define SEMAPHORE_LOCK  0
#define SPIN_LOCK       1
#define SHARED_LOCK     2

int useulocks               = 0;
int lock_type               = SEMAPHORE_LOCK;
int threaded_model          = 0;
int shared_ratio            = 0;
unsigned long spin_patience = 0;

int init_locks(ulock_t* ulocks[], int numlocks, int shared, 
	       int memtype, unsigned int nummobjs, int *lockimpl);
int release_locks(void);

#define SHARED_DATA_MEMOBJ  (MAX_MEMOBJS)

static struct mobjdescr {
	void *addr;
	int   size;
	int   id;
	int   type;
} mem_objects[MAX_MEMOBJS+1];

void *get_shared_memory(int size, int type, int idx);
void release_shared_memory();


int P_lock(ulock_t*);
int V_lock(ulock_t*);
int readlock(ulock_t*);
int readunlock(ulock_t*);
int writelock(ulock_t*);
int writeunlock(ulock_t*);

/* Macros related to active set formation 
 * depend on setsize,num_active,setid 
 * setid is a per-thread value that should be defined properly
 * before macros depending on it are called 
 */

#define isextra(myid)   (myid >= (regsetsize * num_active))      
#define idextra(setid)  ((regsetsize * num_active) + setid)
#define needextra(setid) (setid < num_children - (regsetsize*num_active))

#define regsetstart(setid)  (regsetsize * setid)
#define regsetend(setid)    ((regsetsize * (setid+1)) - 1)


/*
 * system calls
 */

int __clone (int (*fn) (void *arg), void *thread_stack, int flags, void *arg);

/*
 * prototypes for this file.
 */

#define OUTPUT_NO_CONF_FAILURE       (0x04)
#define OUTPUT_SHOW_INDIVIDUAL_LOCKS (0x08)
#define OUTPUT_LOCK_STATUS           (0x10)

void run_test_time(void); 
int child_code(void *arg);

char *progname;
int num_children = 200;	        /* number of child processes to create */
int num_active = 1; 
int num_seconds = 20 ;          /* number of seconds to run test */
int foutput = 0;                /* controls type of output */
int verbose = 0;                 /* controls type of output */
int rnd_compute_time = 300;     /* [ 0 .. # ] microseconds of computation time */
int rnd_hold_time    = 300;     /* [ 0 .. # ] microseconds of lock hold time */

int start_sem;			/* sem to serialize test */
int valid_test = 1;
double rounds_per_usec = 0.0 ; /* obtained through calibration */
int local_exec_count = 0;       /* unused */
int max_trials = MAX_TRIALS;
int min_trials;
int calibration_only = 0;

float tau[MAX_TRIALS-1] = { 12.706, 4.303, 3.182, 2.776, 2.571, 2.447, 2.365, 2.306, 2.262, 2.228, 2.201, 2.179, 2.160, 2.145, 2.131, 2.120, 2.110, 2.101, 2.093, 2.086, 2.080, 2.074, 2.069, 2.064 };

struct _results
{
	float data;
	float ave;
	float var;
	float conf;
} results[MAX_TRIALS] = {{0.0}};

/******************** Statistical functions **********************************/

double probrange(unsigned long top)
{
	double value = random();
	value = (top*value) / ((double)RAND_MAX);
	return value;
}

float variance(int n, float sum, float sum2)
{
	return ((((float)n * sum2)-(sum * sum))/((float)n * (float)(n - 1)));
}

/*
 * iterertions includes NUM_WARMUP + a min of 3 iterations
 */
int confidence(int iter)
{
	float	sum_trials  = 0.0;
	float	sum_trials2 = 0.0;
	float	percent     = 0.01;
	int	i,x,y;
	
	x = iter - NUM_WARMUP;
	y = iter - 1;

	/* compute average */

	if (iter <= NUM_WARMUP) {
		return(0);
	} else {
 		for (i = NUM_WARMUP ; i < iter ; i++) {
			sum_trials  += results[i].data;
			sum_trials2 += results[i].data * results[i].data;
		}
	}
	results[y].ave = sum_trials / (float)(x);

#if 0
	if (iter < NUM_WARMUP + 2)
		return(0);
#endif

	/* compute the variance */

	results[y].var = variance(x,sum_trials,sum_trials2);
	if (isnan(results[y].var))
		results[y].var = 0.0;
	if (results[y].var < 0.0)
		results[y].var = 0.0;

	/* 95% confident that ave is within percent% of "true" average ? */

	results[y].conf = tau[x-2] * sqrt(results[y].var / (float)x);
	if (isnan(results[y].conf))
		results[y].conf = 0.0;
	if (results[y].var < 0.0)
		results[y].var = 0.0;

	if (iter < min_trials)
		return(0);

	if (results[y].conf <= results[y].ave * percent) return (1);
	
	return(0);

}

void child_statistics(double *mean, double *variance) 
{
	int i;
	double sumdiff=0.0 ;
        double sum_trials=0.0;

	for (i = 0 ; i < num_children ; i++) 
		sum_trials  += (double) results[i].data;
	*mean =  (sum_trials/(double)num_children);

	for (i = 0 ; i < num_children ; i++) {
		double meandiff = (double)results[i].data - *mean ; 
		sumdiff += meandiff*meandiff;
	}
	
	*variance = (num_children > 1) ? (sumdiff/((double)num_children - 1)) : 0.0;
}

double uniform(double mean)
{
	/* Generate a random uniformly in [0.5xmean .. 1.5xmean] */

	return ( (mean == 0.0) 
		 ? 0.0 
		 : (random()/(double)RAND_MAX + 0.5) * mean );
}

double uniform2(double mean)
{
	/* Generate a random uniformly in [0..mean] */

	double value;
	return ( (mean == 0.0) 
		 ? 0.0 
		 : (random()/(double)RAND_MAX) * mean );
}
	
double local_exec()
{
	unsigned int a = 0, b=0;
	memcpy(&a,&b,1);
}

double calibration(int secs)
{
	/* figure out how many loops we can execute per micro second */
	int i;
	int count = 0;
	unsigned long n_initial = 100000;
	unsigned long clock1, clock2, clockterm;

	clock1 = clock() ;
	clockterm = clock1 + secs*CLOCKS_PER_SEC;
	do {
		for(i=0; i< n_initial; i++)
		{
			local_exec();
		}
		clock2 = clock();
		count++;
	} while (clock2 < clockterm);

	n_initial *= count;

	return ((double)n_initial*CLOCKS_PER_SEC) / ((double)(clock2-clock1) * 1000000.0);
}


void sighandler(int sig)
{
        shared->stop_test = 3;
}

void usage()
{

	printf ("Usage: %s [ <args> ] \n\n",progname);
	printf ("version 1.0.0  frankeh@watson.ibm.com  < (C) 2001 > \n\n");
	printf ("program to test user level locks over a wide variety of "
		"test scenarios\n\n");

	printf ("-a N, number of concurrent locks.\n");
	printf ("-c N, number of child processes to create\n");
	printf ("-t N, number of seconds to run test before checking throughput\n");
	printf ("-r N, number of usecs of computation lock NOT held\n");
	printf ("-x N, number of usecs of computation lock held\n\n");
	printf ("-S  , regular semaphores (no spinning)  [ default ]]\n");
	printf ("-p N, spin locks with N usecs patience\n\n");
	printf ("-s N, shared locks with N\% read requests, rest write\n");
	printf ("-M c, memory type (s=shared memory, m=file)   default=s\n");
	printf ("-C    calibration only return used for \"-R\" arg\n");
	printf ("-R N, use previously reported calibration factor\n");
	printf ("-m N, number of maxtrials before stopping confidence test\n");
	printf ("-T  , threaded mode");
	printf ("-P  , process  mode  [ default ]");
	printf ("-v  , verbose\n");
	printf ("-o N, how to report results.\n");
	printf ("      0 iteration tabular output.\n");
	printf ("      1 one-liner\n");
	printf ("      2 for -t, prints 'children , seconds , microseconds/round'\n");
}

int main(int argc, char *argv[])
{
	int c;
	int clone_err;
	int i,rc;
	int exit_rc = 0;
	int wait_status;
	int (*worker) (void *arg);
	struct sched_param param;
	struct rlimit myrlimit;
	char *child_stack ;
	int memtype = 's';
	unsigned int nummobjs = 1;

	progname = argv[0];

	while ((c = getopt(argc, argv, "a:c:m:o:r:s:p:t:vx:z:TPR:CMhH")) != -1) {
	  switch (c) {
	  case 'a': num_active = atoi(optarg); break;
	  case 'c': num_children = atoi(optarg); break;
	  case 'm': max_trials = atoi(optarg); break;
	  case 'o': foutput = atoi(optarg); break;
	  case 'r': rnd_compute_time = atoi(optarg); break; 	  
	  case 's': lock_type = SHARED_LOCK; 
		    shared_ratio = atoi(optarg); break;
	  case 'p': lock_type = SPIN_LOCK;
		    spin_patience = atoi(optarg); break;
	  case 't': num_seconds = atoi(optarg); break;
	  case 'v': verbose = 1; break;
	  case 'x': rnd_hold_time = atoi(optarg); break; 	  
	  case 'z': nummobjs = atoi(optarg); break;
	  case 'R': rounds_per_usec = atof(optarg); break;
	  case 'P': threaded_model = 0; break;
	  case 'T': threaded_model = 1; break;
	  case 'C': calibration_only = 1; break;
	  case 'M': memtype = optarg[0]; break;
	  default:  
		  printf("Unknown option <%c>\n\n",c);
	  case 'h': case 'H':  
	    usage();
	    goto exit_main2;
	  }
	}

        /* calibrate internal loops */
	if (calibration_only) {
		rounds_per_usec = calibration(5);
		printf("Calibration factor = %d\n",(int)rounds_per_usec);
		exit(71);
	}
	if (rounds_per_usec == 0.0)
		rounds_per_usec = calibration(2);

#if 0
	printf("c:%d, t:%d, a:%d, r:%d, o:%d, s:%d\n",
			num_children, num_seconds, num_active, 
			rnd_compute_time,foutput,  shared_ratio);
#endif

	if ((num_seconds < 0) ||
	    (num_children <= 0) || (num_children > MAX_CHILDREN) ||
	    (num_active <= 0) || (num_active > num_children) ||
	    (rnd_compute_time < 0) || (rnd_hold_time < 0)
	   )
	    
	{
		printf("num_seconds:%d, num_children:%d, num_active:%d, rnd_compute_time:%d\n",
		       num_seconds, num_children, num_active, rnd_compute_time);
	    	usage();
		goto exit_main2;
	}
	

	if (max_trials < 0) max_trials *= -1, num_warmup = 0;
	min_trials = (MIN_TRIALS > max_trials) ? max_trials : MIN_TRIALS;

	/* Increase limits on number of open files */
	/* normally 1024 (cur & max), set to MAX_CHILDREN */

	myrlimit.rlim_cur = myrlimit.rlim_max = MAX_CHILDREN*2 ; 

	if (setrlimit(RLIMIT_NOFILE,&myrlimit) != 0) {
		exit_rc = errno ;
		perror("setrlimit() ");
		goto exit_main2;
	}
	
	/* allocate childrens stacks*/
	
	child_stack = malloc(num_children*STACK_SIZE);
	if (child_stack == NULL) {
		exit_rc = errno;
		perror ("malloc of 'child_stack' failed ");
		goto exit_main2;
	}
	
	/* start_sem is used to start all children at same time */	

	start_sem = semget (IPC_PRIVATE, 1, IPC_CREAT | IPC_EXCL | 0660);
	if (start_sem == -1) {
		exit_rc = errno;
		perror("semget(start_sem) IPC_CREATE ");
		goto exit_main4;
	}

	rc = init_locks(ulocks,num_active,(lock_type == SHARED_LOCK),
			memtype,nummobjs,&useulocks);
	if (rc == -1) {
		exit_rc = errno;
		perror("failed to initialize locking subsystem ");
		goto exit_main4;
	}

	shared = get_shared_memory(sizeof(struct shared_data),'s',SHARED_DATA_MEMOBJ);
	if (shared == NULL) {
		exit_rc = errno;
		perror("failed to initialize locking subsystem ");
		goto exit_main4;
	}
	for (i = 0 ; i < num_children ; i++)
		shared->per_child[i].count = 0;

	/* Launch threads */
	
  	worker = child_code;
	for (i=0; i< num_children; i++) {
		clone_err = __clone(worker, 
				    &child_stack[(i+1)*STACK_SIZE],
				    threaded_model ? CLONE_FLAGS_THREAD
				    : CLONE_FLAGS_PROCESS,
				    (void*)i);
		if (clone_err == -1) {
			exit_rc = errno;
			perror ("clone() ");
			goto exit_main5;
		}
	}

	i = 0;
 wait_for_children:
	do {
		if (!shared->per_child[i].pid) {
			usleep(1000);
			goto wait_for_children;
		}
		i++;
	} while (i< num_children);
	
	sleep(1);
	/* Increase priority of parent thread */

	param.sched_priority = 90;
	rc = sched_setscheduler(getpid(), SCHED_FIFO, &param);
	if (rc == -1) {
		exit_rc = errno;
		perror ("sched_setscheduler() ");
		goto exit_main5;
	}

	signal(SIGINT,sighandler);

	run_test_time();
	
exit_main5: 
	/* wait untoil all children complete */	
	for (i = 0 ; i < num_children ; i++) {
		rc = waitpid (-1, &wait_status, __WCLONE);
		if (rc == -1) {
			exit_rc = errno;
			perror ("waitpid() ");
		}
		if (verbose) 
			printf("Child %d returned\n",rc);
	}

exit_main4:
 
	rc = semctl(start_sem, 0, IPC_RMID, 0);
	release_locks();
	release_shared_memory();

exit_main3:

	free(child_stack);

exit_main2:
	return (exit_rc) ;
}

/*************************************************************************
 *  Validity Testing Interface: atomically set a variable
 * test later for modifications
 *************************************************************************/

int show_history(int start)
{
	int i,end;
	int last;
	int same_idx;

	end = atomic_read(&shared->hcnt);
	if (start < 0) start = 0;
	if ((end - HISTORY_COUNT) > start) 
		start -= HISTORY_COUNT;
	
	last = shared->history[HISTORY_IDX(start)];
	same_idx = start;

	for (i=0;i<num_children;i++) 
		printf("child_%d: %d\n",i,shared->per_child[i].pid) ;
	printf("\n");

	for (i=start+1;i<end;i++) {
		int val = shared->history[HISTORY_IDX(start)];
		if (val == last) 
			continue;
		if (same_idx == i-1)
			printf("%4d       :  %4d\n",same_idx,last);
		else
			printf("%4d - %4d:  %4d\n",same_idx,i-1,last);
		last = val;
		same_idx = i;
	}
	printf("%4d - %4d:  %4d\n",same_idx,end-1,last);
}

STATIC_INLINE
void record(int myid)
{
	int i;
	atomic_inc(&shared->hcnt);
	i = atomic_read(&shared->hcnt);
	shared->history[HISTORY_IDX(i)] = myid;
}

STATIC_INLINE 
unsigned long verify_get(int setid, int id) 
{ 
	atomic_t *vptr = &shared->per_lock[setid].verify_values;
	return atomic_read(vptr); 
}

STATIC_INLINE 
unsigned long verify_inc(int setid, int id)
{
	unsigned long val;
	atomic_t *vptr = &shared->per_lock[setid].verify_values;
	atomic_inc(vptr);
	val = atomic_read(vptr);
	return val;
}

STATIC_INLINE 
void verify_test_int(int setid, unsigned long val, int id, int line)
{
	atomic_t *vptr = &shared->per_lock[setid].verify_values;
	unsigned long gval = atomic_read(vptr);
	if (gval != val) {
		yyprintf("%d(%d): BROKEN L%d:  g=%d p=%d s=%d\n",
			 id,getpid(),line,gval,val,setid);
		shared->stop_test = 2;
		//show_history(ulocks[setid]->counters[4]-80);
	}
}

#define verify_test(set,val,id) verify_test_int(set,val,id,__LINE__)

/*************************************************************************
 * Code execute by each spawned child thread 
 *************************************************************************/

int child_code(void *arg)
{
	int i=0, rc, exit_rc = 0;
	int myid;
	struct sembuf mysembuf;
	int msgsize ;
	int regsetsize,setid ;
	unsigned int val;
	int seqn;
	int isread;

	/* Active set formation : 
	 * each id put into sets of size "regsetsize" first
	 * leftover ids distributed amongst regular sets, one per set
	 * e.g.  to divide [0..10] into num_active=3,
	 *       form regular sets : [0,1,2] [3,4,5] [6,7,8] 9,10
	 *       9,10 are "extras", distribute one each to regular sets 
	 *	          forming    [0,1,2,9] [3,4,5,10] [6,7,8]
	 *	 and adjust previd,nextid appropriately	to form logically 
	 *       circular linked lists per set such as (for first set)
	 *                 0 <--> 1 <--> 2 <--> 9 <--> 0 
	 */


	/* Defaults */
	myid = (int) arg ;

	bprintf("Started %d\n",myid);
	/* Determine extras */
	regsetsize = num_children/num_active ;

	if (isextra(myid))
		setid = myid - (regsetsize * num_active) ;
	else
		setid = myid / regsetsize ;
	
	shared->per_child[myid].pid = getpid();
	/* wait to be released */
	mysembuf.sem_num = 0;
	mysembuf.sem_op  = -1;
	mysembuf.sem_flg = 0;
	
	rc = semop(start_sem, &mysembuf, 1);
	if (rc == -1) {
		exit_rc = errno;
		perror ("child semop(start_sem) failed");
		return(exit_rc);
	} 

	/* Actions to be done by each thread */

	while (!shared->stop_test) {
		int comp_unlocked_rounds ;
		int comp_locked_rounds    ;
		comp_unlocked_rounds = (int) (uniform(rnd_compute_time) * rounds_per_usec) ;
		comp_locked_rounds = (int) (uniform(rnd_hold_time) * rounds_per_usec) ;

                /* unlocked */
		for (i=0;i<comp_unlocked_rounds;i++) 
			local_exec();

		/* locked */
		switch (lock_type) {
		case SHARED_LOCK:
			isread = (((unsigned long)(uniform2(100)))<shared_ratio);
			if (isread) {
				if (readlock(ulocks[setid])) 
					goto bad_state_lock;
				val = verify_get(setid,myid);
				for (i=0;i<comp_locked_rounds;i++) 
					local_exec();
				verify_test(setid,val,myid);
				if (readunlock(ulocks[setid]))
					goto bad_state_unlock;
			} else {
				if (writelock(ulocks[setid]))
					goto bad_state_lock;
				val = verify_inc(setid,myid);
				for (i=0;i<comp_locked_rounds;i++) 
					local_exec();
				verify_test(setid,val,myid);
				if (writeunlock(ulocks[setid]))
					goto bad_state_unlock;
			}
			break;

		case SEMAPHORE_LOCK:

			if (P_lock(ulocks[setid])) 
				goto bad_state_lock;
			val = verify_inc(setid,myid);
			for (i=0;i<comp_locked_rounds;i++) 
				local_exec();
			verify_test(setid,val,myid);
			if (V_lock(ulocks[setid]))
				goto bad_state_unlock;
			break;

		case SPIN_LOCK:

			if (P_lock_spin(ulocks[setid],spin_patience)) 
				goto bad_state_lock;
			val = verify_inc(setid,myid);
			for (i=0;i<comp_locked_rounds;i++) 
				local_exec();
			verify_test(setid,val,myid);
			if (V_lock(ulocks[setid]))
				goto bad_state_unlock;
			break;

		}
		/* Record statistic */
		shared->per_child[myid].count++;
 	}
	bprintf("%d ->%d\n",myid,shared->per_child[myid].count);
  	sleep(1);

	return(exit_rc);

 bad_state_lock:
	printf("%d: bad state lock <%s>\n",myid,strerror(errno));
	goto out;

 bad_state_unlock:
	printf("%d: bad state unlock <%s>\n",myid,strerror(errno));

 out:
	exit_rc = errno;
	while (!shared->stop_test) 
		sleep(1);
	return(exit_rc);
	
}

void lock_contention(ulock_t* ulocks[], int numlocks, double results[3])
{
	int i;
	unsigned long w_ok      = 0;
	unsigned long w_spin_ok = 0;
	unsigned long r_ok      = 0;
	unsigned long w_failed  = 0;
	unsigned long r_failed  = 0;
	unsigned long sum, sum1;
	
	for ( i=0 ; i < numlocks ; i++) {
		unsigned long *cnts = ulocks[i]->counters;
		
		w_ok      += cnts[ULOCK_STAT_WRITE_SUCCESS];
		w_spin_ok += cnts[ULOCK_STAT_WRITE_SPIN_SUCCESS];
		w_failed  += cnts[ULOCK_STAT_WRITE_FAILED];
		r_ok      += cnts[ULOCK_STAT_READ_SUCCESS];
		r_failed  += cnts[ULOCK_STAT_READ_FAILED];

		if (foutput & OUTPUT_SHOW_INDIVIDUAL_LOCKS) {
			unsigned long sum;
			double w_cont, r_cont, s_ratio;
			
			sum1= cnts[ULOCK_STAT_WRITE_SPIN_SUCCESS] +
			      cnts[ULOCK_STAT_WRITE_FAILED];
			sum = cnts[ULOCK_STAT_WRITE_SUCCESS] + sum1;

			w_cont =
				(sum == 0) ? 0.0
				: 100.0 * ((double)sum1/ (double) sum);

			sum = cnts[ULOCK_STAT_READ_SUCCESS] +
				cnts[ULOCK_STAT_READ_FAILED];
			r_cont =
				(sum == 0) ? 0.0
				: 100.0 * ((double)cnts[ULOCK_STAT_READ_FAILED] / (double) sum);
			s_ratio = 
			        (sum1 == 0) ? 0.0
			        : 100.0 * ((double)cnts[ULOCK_STAT_WRITE_SPIN_SUCCESS]/(double)sum1);
			printf("%p %ld %ld %ld %ld %ld %04.2lf %04.2lf %4.0lf\n",
			       ulocks[i],
			       cnts[ULOCK_STAT_WRITE_SUCCESS],
			       cnts[ULOCK_STAT_WRITE_FAILED],
			       cnts[ULOCK_STAT_READ_SUCCESS],
			       cnts[ULOCK_STAT_READ_FAILED],
			       cnts[ULOCK_STAT_WRITE_SPIN_SUCCESS],
			       w_cont, r_cont,s_ratio);
		}
	}
	sum = (w_ok + w_spin_ok + w_failed);
	results[0] = (sum == 0) ? 0.0 : ((double)w_failed / (double) sum);
	sum = (r_ok + r_failed);
	results[1] = (sum == 0) ? 0.0 : ((double)r_failed / (double) sum);
	sum = (w_ok + w_spin_ok);
	results[2] = (sum == 0) ? 0.0  : ((double)w_spin_ok / (double) sum);

	if (foutput & OUTPUT_SHOW_INDIVIDUAL_LOCKS)
		printf("#=%d s=%ld f=%ld rs=%ld rf=%ld ss=%ld wc=%4.2lf%% rc=%4.2lf%% sr=%4.2lf%%\n",
		       numlocks,w_ok,w_failed,r_ok,r_failed,w_spin_ok,
		       100.0*results[0], 100.0*results[1],100.0*results[2]);
}

void run_test_time(void)
{
	int i,rc;
	int exit_rc = 0;
	int iterations = 0;
	unsigned long long x;
	unsigned long long t, prev_t;
	struct sembuf mysembuf1;
	double total_avg;
	double contention[3];
	struct timezone tz1,tz2;
	struct timeval  tv1,tv2,tvr;
	int reached_confidence = 0;

	mysembuf1.sem_num = 0;
	mysembuf1.sem_op  = num_children;
	mysembuf1.sem_flg = 0;
	
	/* post the sema4 to allow children to start */

	rc = semop (start_sem, &mysembuf1, 1);
	if (rc == -1) {
		shared->stop_test = 1;
		exit_rc = errno;
		perror ("parent semop(start_sem) failed ");
		goto exit_test;
	}

restart:

        prev_t = 0;
        t = 0;
 
        /* get the start time */
 
        rc = gettimeofday (&tv1, &tz1);
        if (rc) {
                shared->stop_test = 1;
                exit_rc = errno;
                perror ("gettimeofday failed on tv1 ");
                goto exit_test;
        }
 
        /* wait until timeout */
 
        for (i = 0 ; i < num_children ; i++) {
		prev_t += shared->per_child[i].count;
	}

        sleep (num_seconds);

        for (i = 0 ; i < num_children ; i++) {
		t += shared->per_child[i].count;
	}
 
        /* get end time */
 
        rc = gettimeofday (&tv2, &tz2);
        if (rc) {
                shared->stop_test = 1;                
                exit_rc = rc;
                perror ("gettimeofday failed on tv2 ");
                goto exit_test;
        }
 
	yyflush(1);  /* flush output buffer */

        /* compute microseconds per transfer */
 
        timersub(&tv2, &tv1, &tvr); /* tvr now contains result of tv2-tv1 */
 
        x = (unsigned long long)tvr.tv_sec * 1000;
        x+= (unsigned long long)tvr.tv_usec/1000;
        results[iterations].data = (float)(t - prev_t);
        results[iterations].data /= (float)x;
	{
		double mean;
		double variance;
		child_statistics(&mean,&variance);
		bprintf("Rerun : Across children : Avg %15.2f \t Var %15.2f\n",
			mean,variance);
	}
 
        iterations++;
	if (!shared->stop_test) {
		if (reached_confidence = confidence(iterations)) 
			shared->stop_test = 1;
		else if (iterations >= max_trials)
			shared->stop_test = 1;
		else
			goto restart;
	}         
	if (verbose) printf (" Test Completed %d.\n",shared->stop_test);

	total_avg = 0.0;
	for (i = NUM_WARMUP ; i < iterations ; i++) {
		total_avg += results[i].data;
	}
	total_avg /= (iterations-NUM_WARMUP);
	total_avg *= 1000;
	lock_contention(ulocks,num_active,contention);

	
	if (foutput & OUTPUT_NO_CONF_FAILURE) {
		foutput &= ~OUTPUT_NO_CONF_FAILURE;
		valid_test = 1;
	} else {
		if (!reached_confidence) 
			printf("\n*****> failed to reach confidence level <*****\n");
	}
	switch (foutput & 0x3) {
		
	case 0:
		printf ("| OPs/sec | |   ave   | | variance| |confidence|\n");
		printf ("----------- ----------- ----------- ------------\n");
		for (i = 0 ; i < iterations ; i++) {
			printf ("%11.3f %11.3f %11.3f %11.3f ",
				results[i].data,
				results[i].ave,
				results[i].var,
				results[i].conf);
			if (i < NUM_WARMUP)
				printf ("warmup");
			printf ("\n");
		}

	case 1:
		if (!valid_test) break;
		printf ("%2u %2u %2u %2d %c %3d %3d %3d %3d  %6.2lf %6.2lf %6.2lf %8.0lf\n", 
			num_children, num_active,
			num_seconds, iterations - num_warmup,
			useulocks ? 'u' : 'k',
			rnd_compute_time, rnd_hold_time,
			(lock_type == SHARED_LOCK)?shared_ratio:-1,
			(lock_type == SPIN_LOCK)?spin_patience:-1,
			100*contention[0], 
			100*contention[1], 
			100*contention[2], 
			total_avg);
		break;
		
	case 2:
		if (!valid_test) break;
		printf ("%u , %u , %u , %0.3f\n", 
			num_children, num_active, num_seconds, total_avg);
		break;
		
	default:
		if (!valid_test) break;
		printf ("%u children performed %ld %9.0lf iterations \n", 
			num_children, (unsigned long)total_avg, total_avg);
		break;
	}

	if (foutput & OUTPUT_LOCK_STATUS) {
		for ( i=0 ; i < num_active ; i++) {
			printf("lock_%d: %lx\n",i,ulocks[i]->status);
		}
	}
exit_test:	
	return;	
}


/**********************************************************************
 *
 * Dealing with shared memory regions
 *    by default the shared_data is located in the SHARED_DATA_MEMOBJ 
 *    region which is allocated as a shared memory segment
 *    the lock data region can be allocated either over a shared file
 *
 **********************************************************************/


#define PAGE_SIZE (4096)
#define PAGE_ALIGN(x) (((x) + PAGE_SIZE - 1) & ~(PAGE_SIZE-1))

void *get_shared_memory(int size, int type, int idx)
{
	int id,i;
	key_t key;
	void* maddr;

	size = PAGE_ALIGN(size);
	switch (type) {
	case 'f':
		if ( (id = open("./ulockflex.map",O_CREAT|O_RDWR)) == -1) {
			return NULL;
		}
		maddr = (void*) mmap(NULL,size,
			     PROT_WRITE | PROT_READ, /* allow read and write */
			     MAP_SHARED | MAP_SEMAPHORE, /* make it shared       */
			     id, 0);
		break;
	case 's':
		key = ftok("./ulockflex.tok",idx);
		if (key == -1)
			return NULL;
		id = shmget(key,size,IPC_CREAT );
		if (id == -1) 
			return NULL;
		maddr = shmat(id,NULL,0);  
		if (maddr == ((void*)-1))
			return NULL;
		/* now mark the shm-seg for destroy, 
		 * actually only after last detach will it be deleted
		 */
		shmctl(id,IPC_RMID,NULL);
		break;
	default:
		return NULL;
	}
	mem_objects[idx].addr = maddr;
	mem_objects[idx].size = size;
	mem_objects[idx].id   = id;
	mem_objects[idx].type = type;
	return maddr;
}

void release_shared_memory()
{
	int i;

	for (i=0; i < MAX_MEMOBJS+1; i++) {
		if (mem_objects[i].addr == NULL) continue;
		if (mem_objects[i].type == 'f')
			munmap(mem_objects[i].addr,mem_objects[i].size);
		else
			shmdt(mem_objects[i].addr);
	}
} 

